unit MGSCore;

interface

  uses SysUtils, Classes, Utils, MGSApi, Dialogs;

  type
    mgsCompareOperator = (mgsCT_Great, mgsCT_Less, mgsCT_GreatEqual, mgsCT_LessEqual, mgsCT_Equal, mgsCT_NonEqual);
    mgsMathOperator =
        (mgsMT_Plus, mgsMT_Minus, mgsMT_Mul, mgsMT_Div,
         mgsMT_And, MgsMT_Or, MgsMT_Not,
         mgsMT_Great, mgsMT_Less, mgsMT_GreatEqual, mgsMT_LessEqual, mgsMT_Equal, mgsMT_NonEqual);
    mgsBoolOperator = (mgsBT_And, MgsBT_Or);
    TMGSVariable = class(TInterfacedObject,IMGSVariable)
      private
        _Handle: cardinal;
        _Type: mgsVarType;
        _Name: PAnsiChar;
        _Value: PAnsiChar;
        _IsConstant: boolean;
        procedure _SetValue (value: PAnsiChar);
      public
        //=============================
        //  Interface methods
        //===
        function GetHandle: cardinal;
        function GetName: PAnsiChar;
        function GetValue: PAnsiChar;
        procedure SetValue( value: PAnsiChar );
        function GetType: mgsVarType;
        function IsConstant: boolean;
        //=============================
        property Handle: cardinal read _Handle;
        property VariableType: mgsVarType read _Type;
        property Value: PAnsiChar read _Value write _SetValue;
        property Name: PAnsiChar read _Name;
      published
        constructor Create(Name: PAnsiChar; VType: mgsVarType; Value: PAnsiChar; isConstant: boolean);
        constructor CreateFrom(OtherVariable: IMGSVariable);
        destructor Destroy; override;
    end;
    TMGSVariableScope = class(TInterfacedObject, IMGSVariableScope)
      private
        _HandleStack: TIntStack;
        _VariableList: array of IMGSVariable;
        _VariableCount: cardinal;
        _Name: PAnsiChar;
      public
        //=========================
        //  Interface methods
        //===
        function GetName: PAnsiChar;
        function AddNewVariable(Name: PAnsiChar; VType: mgsVarType; Value: PAnsiChar; IsConstant: boolean; out Variable: IMGSVariable): cardinal;
        function GetVariable(Handle: cardinal; out Variable: IMGSVariable): cardinal; overload;
        function GetVariable(Name: PAnsiChar; out Variable: IMGSVariable): cardinal; overload;
        function FreeVariable(Handle: cardinal): cardinal;
      published
        constructor Create(Name: PAnsiChar);
        destructor Destroy; override;
    end;
    TMGSFunction = class(TInterfacedObject, IMGSFunction)
      private
        _Handle: cardinal;
        _VariableScope: IMGSVariableScope;
        _IsNative: boolean;
        _Name: PAnsiChar;
        _Type: mgsFuncType;
        //For custom function
        _FunctionCode: IMGSScript;
        //For native function
        _NativeFuncAddress: Pointer;
        _ParameterList: TMGSFuncParameterList;
        _ResultType: mgsVarType;
      public
        function GetType: mgsFuncType;
        function GetName: PAnsiChar;
        function GetHandle: cardinal;
        function VariableScope: IMGSVariableScope;
        function GetReturnType: mgsVarType;
        function Execute (Values: TMGSFuncValueList; var returnValue: PAnsiChar): cardinal;
      public
        constructor Create (Name: PAnsiChar; FType: mgsFuncType; ParameterList: TMGSFuncParameterList; Count: integer; FunctionCode: PAnsiChar; ParentFunctionScope: IMGSFunctionScope); overload;
      published
        constructor Create (Name: PAnsiChar; FType: mgsFuncType; ParameterList: TMGSFuncParameterList; Count: integer; NativeFuncAddress: Pointer); overload;
        destructor Destroy; override;
    end;
    TMGSFunctionScope = class(TInterfacedObject, IMGSFunctionScope)
      private
        _FunctionList: array of IMGSFunction;
        _FunctionCount: cardinal;
        _HandleStack: TIntStack;
        _Name: PAnsiChar;
      public
        function GetName: PAnsiChar;
        function AddNewFunction(Name: PAnsiChar; FType: mgsFuncType; ParameterList: TMGSFuncParameterList; ParameterCount: integer; Address: Pointer; out Func: IMGSFunction): cardinal; overload;
        function AddNewFunction(Name: PAnsiChar; FType: mgsFuncType; ParameterList: TMGSFuncParameterList; ParameterCount: integer; FunctionCode: PAnsiChar; ParentFunctionScope: IInterface; out Func: IMGSFunction): cardinal; overload;
        function GetFunction(Handle: cardinal; out Func: IMGSFunction): cardinal; overload;
        function GetFunction(Name: PAnsiChar; out Func:  IMGSFunction): cardinal; overload;
      published
        constructor Create(Name: PAnsiChar);
        destructor Destroy; override;
    end;

    TMGSTokenType = (mgsTT_keyword, mgsTT_operator, mgsTT_name, mgsTT_delimeter, mgsTT_intconst, mgsTT_strconst, mgsTT_boolconst);

    TMGSToken = record
      _text: string;
      _type: TMGSTokenType;
    end;

    TMGSScript = class(TInterfacedObject, IMGSScript)
      private
        _TokenList: array of TMGSToken;
        _TokenCount: integer;

        _CodeList: TStrings;
        _Name: PAnsiChar;
        _VariableScope: IMGSVariableScope;
        _FunctionScope: IMGSFunctionScope;

        procedure ParseTokens();

        function IsFunctionDef(pos: integer): boolean;
        procedure FunctionDef(var pos: integer);

        function IsScriptStart(pos: integer): boolean;
        function IsScriptStop(pos: integer): boolean;

        function IsVariableDelete(pos: integer): boolean;
        procedure VariableDelete(var pos: integer);

        function IsFunctionCall(pos: integer): boolean;
        function FunctionCall(var pos: integer): TMGSValue;

        function IsVariableDef(pos: integer): boolean;
        procedure VariableDef(var pos: integer);

        function IsLoop(pos: integer): boolean;
        function IsEndLoop(pos: integer): boolean;
        function IsExitWhen(pos: integer): boolean;
        function LoopExitWhen(var pos: integer): boolean;

        function IsIF(pos: integer): boolean;
        function IsElse(pos: integer): boolean;
        function IsEndIF(pos: integer): boolean;
        function IfAction(var pos: integer): boolean;

        function IsAssignOperation(pos: integer): boolean;
        procedure AssignOperation(var pos: integer);

        function GetValue(var pos: integer; subValue: boolean): TMGSValue;
        procedure SkipSpaceTokens(var pos: integer);
        
      public
        function GetName: PAnsiChar;
        procedure Load(text: PAnsiChar);
        procedure Execute();
      public
        constructor Create(Name: PAnsiChar; FunctionScope: IMGSFunctionScope; VariableScope: IMGSVariableScope); overload;
      published
        constructor Create(Name: PAnsiChar); overload;
        destructor Destroy(); override;
    end;

    PMathTree = ^TMathTree;
    TMathTreeType = (mathTT_value, mathTT_operator, mathTT_sub, mathTT_result);
    TMathTree = record
      _Parent: PMathTree;
      _Left: PMathTree;
      _Right: PMathTree;
      _Type: TMathTreeType;
      _TokenPos: integer;
      _Value: TMGSValue;
      _Text: string;
    end;


    function GetVarType(s: string): mgsVarType;

    procedure Init;

  var
    GlobalFunctionScope : IMGSFunctionScope;
    GlobalVariableScope : IMGSVariableScope;


implementation

  const
    defIntValue : PAnsiChar = '0';
    defRealValue: PAnsiChar = '0.0';
    defStringValue: PAnsiChar = '';
    defBooleanValue: PAnsiChar = 'False';

    tokentype_delimeter: string = #13#10#32#9+'".,:;(){}[]';
    tokentype_keyword: array [0..18] of string = ( 'int','str','bool','real',
                                                  'def', 'stop',
                                                  'loop', 'endloop', 'exitwhen',
                                                  'if', 'else', 'endif',
                                                  'delete', 'local', 'global',
                                                  'InitEnd', 'Stop',
                                                  'function', 'endfunction'
                                                  );
    keyword_count: integer = Length(tokentype_keyword);
    tokentype_operator: array [0..13] of string = ('=','+','-','*','/','or','and','not','<','>','<=','>=','==', '!=');
    operator_count: integer = Length(tokentype_operator);


  function GetVarType(s: string): mgsVarType;
  begin
    Result := mgsVT_Str;
    if s = 'int' then
      Result := mgsVT_Int
    else if s = 'str' then
      Result := mgsVT_Str
    else if s = 'bool' then
      Result := mgsVT_Boolean
    else if s = 'real' then
      Result := mgsVT_Real;
  end;

  function TMathTree_Draw(base: PMathTree): string;
  begin
    Result := '';

    if base = nil then Exit;

    Result := '('+base._Text + ':';

    if base._Left <> nil then
      Result := Result + TMathTree_Draw(base._Left)
    else
      Result := Result + '~';
    Result := Result + ';';
    if base._Right <> nil then
      Result := Result + TMathTree_Draw(base._Right)
    else
      Result := Result + '~';
    Result := Result + ')';
  end;

  procedure TMathTree_InsertLeft(base, child: PMathTree);
  begin
    child._Left := base._Left;
    base._Left := child;
    if child <> nil then
    begin
      if child._Left <> nil then
        child._Left._Parent := child;
      child._Parent := base;
    end;
  end;

  procedure TMathTree_InsertRight(base, child: PMathTree);
  begin
    child._Left := base._Right;
    base._Right := child;
    if child <> nil then
    begin
      if child._Left <> nil then
        child._Left._Parent := child;
      child._Parent := base;
    end;
  end;

  function TMathTree_Create(): PMathTree;
  begin
    New(Result);
    Result._Parent := nil;
    Result._Left := nil;
    Result._Right := nil;
    Result._Text := '';
    Result._TokenPos := -1;
  end;

  procedure TMathTree_AddChild(base, child: PMathTree; isLeft: boolean);
  begin
    if isLeft then
      base._Left := child
    else
      base._Right := child;
    child._Parent := base;
  end;

  function BoolValues(ValueA, ValueB: TMGSValue; operatorType: mgsBoolOperator; var BoolResult: boolean): cardinal;
    var AType, BType: mgsVarType;
        AVal, BVal: PAnsiChar;
  begin

    AType := MGSValueGetType(@ValueA);
    BType := MGSValueGetType(@ValueB);
    AVal := MGSValueGetValue(@ValueA);
    BVal := MGSValueGetValue(@ValueB);

    Result := MGS_E_OK;

    if (Atype = mgsVT_Boolean) and (BType = mgsVT_Boolean) then
    begin
      case operatorType of
        mgsBT_And:
          BoolResult := StrToBool(AVal) and StrToBool(BVal);
        mgsBT_Or:
          BoolResult := StrToBool(AVal) or StrToBool(BVal);
      end;
    end else
      Result := MGS_E_UNCORRECT_TYPE;
  end;

  function MathValues(ValueA, ValueB: PMGSValue; operatorType: mgsMathOperator; MathResult: PMGSValue): cardinal;
    var AType, BType: mgsVarType;
        AVal, BVal: PAnsiChar;
        r: PAnsiChar;
        CompareResult: boolean;
  begin

    AType := MGSValueGetType(ValueA);
    BType := MGSValueGetType(ValueB);
    AVal := MGSValueGetValue(ValueA);
    BVal := MGSValueGetValue(ValueB);

    Result := MGS_E_OK;
    //New(MathResult);
    MathResult.isConst := true;
    if (operatorType < mgsMT_Great) then
      //Std Math
      case AType of
        mgsVT_Int:
          begin
            MathResult._Type := mgsVT_Int;
            if (BType = mgsVT_Int) or (BType = mgsVT_Real) then
            begin
              case operatorType of
                mgsMT_Plus:
                  r := PAnsiChar(IntToStr(FloatStrToInt(AVal) + FloatStrToInt(BVal)));
                mgsMT_Minus:
                  r := PAnsiChar(IntToStr(FloatStrToInt(AVal) - FloatStrToInt(BVal)));
                mgsMT_Mul:
                  r := PAnsiChar(IntToStr(FloatStrToInt(AVal) * FloatStrToInt(BVal)));
                mgsMT_Div:
                  r := PAnsiChar(IntToStr(FloatStrToInt(AVal) div FloatStrToInt(BVal)));
                else
                  begin
                    Result := MGS_E_UNCORRECT_OPERATION;
                    Exit;
                  end;
              end;
            end else
              Result := MGS_E_UNCORRECT_TYPE;
          end;
        mgsVT_Real:
          begin
            MathResult._Type := mgsVT_Real;
            if (BType = mgsVT_Real) or (BType = mgsVT_Int) then
            begin
              case operatorType of
                mgsMT_Plus:
                  r := PAnsiChar(RealToStr(StrToReal(AVal) + StrToReal(BVal)));
                mgsMT_Minus:
                  r := PAnsiChar(RealToStr(StrToReal(AVal) - StrToReal(BVal)));
                mgsMT_Mul:
                  r := PAnsiChar(RealToStr(StrToReal(AVal) * StrToReal(BVal)));
                mgsMT_Div:
                  r := PAnsiChar(RealToStr(StrToReal(AVal) / StrToReal(BVal)));
                else
                  begin
                    Result := MGS_E_UNCORRECT_OPERATION;
                    Exit;
                  end;
              end;
            end else
              Result := MGS_E_UNCORRECT_TYPE;
          end;
        mgsVT_Str:
          begin
            MathResult._Type := mgsVT_Str;
            case operatorType of
              mgsMT_Plus:
                r := StrCat(AVal, BVal);
              else
              begin
                Result := MGS_E_UNCORRECT_OPERATION;
                Exit;
              end;
            end;
          end;
        mgsVT_Boolean:
          begin
            MathResult._Type := mgsVT_Boolean;
            if (BType = mgsVT_Boolean) then
            begin
              case operatorType of
                mgsMT_And:
                  r := PAnsiChar(BoolToStr(StrToBool(AVal) and StrToBool(BVal)));
                mgsMT_Or:
                  r := PAnsiChar(BoolToStr(StrToBool(AVal) or StrToBool(BVal)));
                else
                  begin
                    Result := MGS_E_UNCORRECT_OPERATION;
                    Exit;
                  end;
              end;
            end else begin
              Result := MGS_E_UNCORRECT_TYPE;
            end;
          end;
        else
          if ((operatorType = mgsMT_Not) and (BType = mgsVT_Boolean)) then
          begin
            MathResult._Type := mgsVT_Boolean;
            r := PAnsiChar(BoolToStr(not StrToBool(BVal)));
          end else
          if ((operatorType = mgsMT_Minus) and ((BType = mgsVT_Int) or (BType = mgsVT_Real))) then
          begin
            MathResult._Type := BType;
            if (BType = mgsVT_Int) then
              r := PAnsiChar(IntToStr(-StrToInt(BVal)))
            else
              r := PAnsiChar(RealToStr(-StrToReal(BVal)));
          end else begin
            Result := MGS_E_UNCORRECT_TYPE;
            Exit;
          end;
        end
      else begin
        //Compres math
        MathResult._Type := mgsVT_Boolean;
        CompareResult := false;

        case AType of
          mgsVT_Int,
          mgsVT_Real:
            begin
              if (BType = mgsVT_Int) or (BType = mgsVT_Real) then
              begin
                case operatorType of
                  mgsMT_Great:
                    CompareResult := StrToReal(AVal) > StrToReal(BVal);
                  mgsMT_Less:
                    CompareResult := StrToReal(AVal) < StrToReal(BVal);
                  mgsMT_GreatEqual:
                    CompareResult := StrToReal(AVal) >= StrToReal(BVal);
                  mgsMT_LessEqual:
                    CompareResult := StrToReal(AVal) <= StrToReal(BVal);
                  mgsMT_Equal:
                    CompareResult := StrToReal(AVal) = StrToReal(BVal);
                  mgsMT_NonEqual:
                    CompareResult := StrToReal(AVal) <> StrToReal(BVal);
                end;
              end else begin
                Result := MGS_E_UNCORRECT_TYPE;
                Exit;
              end;
            end;
          mgsVT_Str:
            begin
              if BType = mgsVT_Str then
              begin
                case operatorType of
                  mgsMT_Equal:
                    CompareResult := StrComp(AVal, BVal) = 0;
                  mgsMT_NonEqual:
                    CompareResult := StrComp(AVal, BVal) <> 0;
                  mgsMT_Great:
                    CompareResult := StrComp(AVal, BVal) > 0;
                  mgsMT_Less:
                    CompareResult := StrComp(AVal, BVal) < 0;
                else
                  Result := MGS_E_UNCORRECT_COMPARE;
                end;
              end else begin
                Result := MGS_E_UNCORRECT_TYPE;
                Exit;
              end;
            end;
          mgsVT_Boolean:
            begin
              if BType = mgsVT_Boolean then
              begin
                case operatorType of
                  mgsMT_Equal:
                    CompareResult := StrToBool(AVal) = StrToBool(BVal);
                  mgsMT_NonEqual:
                    CompareResult := StrToBool(AVal) <> StrToBool(BVal);
                else
                  Result := MGS_E_UNCORRECT_COMPARE;
                end;
              end else begin
                Result := MGS_E_UNCORRECT_TYPE;
                Exit;
              end;
            end;
        end;
        r := PAnsiChar(BoolToStr(CompareResult));
      end;
      MathResult._Value := GetMemory(StrLen(r)+1);
      StrMove(MathResult._Value, r, StrLen(r)+1);
   end;

  function CompareValues(ValueA, ValueB: TMGSValue; compareType: mgsCompareOperator; var CompareResult: boolean): cardinal;
    var AType, BType: mgsVarType;
        AVal, BVal: PAnsiChar;
  begin

    AType := MGSValueGetType(@ValueA);
    BType := MGSValueGetType(@ValueB);
    AVal := MGSValueGetValue(@ValueA);
    BVal := MGSValueGetValue(@ValueB);

    Result := MGS_E_OK;
    CompareResult := false;
    case AType of
      mgsVT_Int:
        begin
          if BType = mgsVT_Int then
          begin
            case compareType of
              mgsCT_Great:
                CompareResult := StrToInt(AVal) > StrToInt(BVal);
              mgsCT_Less:
                CompareResult := StrToInt(AVal) < StrToInt(BVal);
              mgsCT_GreatEqual:
                CompareResult := StrToInt(AVal) >= StrToInt(BVal);
              mgsCT_LessEqual:
                CompareResult := StrToInt(AVal) <= StrToInt(BVal);
              mgsCT_Equal:
                CompareResult := StrToInt(AVal) = StrToInt(BVal);
              mgsCT_NonEqual:
                CompareResult := StrToInt(AVal) <> StrToInt(BVal);
            end;
          end else
            Result := MGS_E_UNCORRECT_TYPE;
          Exit;
        end;
      mgsVT_Real:
        begin
          if (BType = mgsVT_Int) or (BType = mgsVT_Real) then
          begin
            case CompareType of
              mgsCT_Great:
                CompareResult := StrToReal(AVal) > StrToReal(BVal);
              mgsCT_Less:
                CompareResult := StrToReal(AVal) < StrToReal(BVal);
              mgsCT_GreatEqual:
                CompareResult := StrToReal(AVal) >= StrToReal(BVal);
              mgsCT_LessEqual:
                CompareResult := StrToReal(AVal) <= StrToReal(BVal);
              mgsCT_Equal:
                CompareResult := StrToReal(AVal) = StrToReal(BVal);
              mgsCT_NonEqual:
                CompareResult := StrToReal(AVal) <> StrToReal(BVal);
            end;
          end else
            Result := MGS_E_UNCORRECT_TYPE;
          Exit;
        end;
      mgsVT_Str:
        begin
          if BType = mgsVT_Str then
          begin
            case CompareType of
              mgsCT_Equal:
                CompareResult := AVal = BVal;
              mgsCT_NonEqual:
                CompareResult := AVal <> BVal;
            else
              Result := MGS_E_UNCORRECT_COMPARE;
            end;
          end else
            Result := MGS_E_UNCORRECT_TYPE;
          Exit;
        end;
      mgsVT_Boolean:
        begin
          if BType = mgsVT_Boolean then
          begin
            case CompareType of
              mgsCT_Equal:
                CompareResult := StrToBool(AVal) = StrToBool(BVal);
              mgsCT_NonEqual:
                CompareResult := StrToBool(AVal) <> StrToBool(BVal);
            else
              Result := MGS_E_UNCORRECT_COMPARE;
            end;
          end else
            Result := MGS_E_UNCORRECT_TYPE;
          Exit;
        end;
    end;
  end;

{ TMGSVariable }

  constructor TMGSVariable.Create(Name: PAnsiChar; VType: mgsVarType; Value: PAnsiChar; IsConstant: boolean);
  begin
    inherited Create;
    _Name := Name;
    _Type := VType;
    _IsConstant := false;
    _SetValue(Value);
    _IsConstant := IsConstant;
  end;

  constructor TMGSVariable.CreateFrom(OtherVariable: IMGSVariable);
  begin
    inherited Create();
    _Name := OtherVariable.GetName();
    _Type := OtherVariable.GetType;
    _SetValue(OtherVariable.GetValue());
  end;

  destructor TMGSVariable.Destroy;
  begin
    inherited Destroy;
  end;

  function TMGSVariable.GetHandle: cardinal;
  begin
    Result := _Handle;
  end;

  function TMGSVariable.GetName: PAnsiChar;
  begin
    Result := PAnsiChar(_Name);
  end;

  function TMGSVariable.GetType: mgsVarType;
  begin
    Result := _Type;
  end;

  function TMGSVariable.GetValue: PAnsiChar;
  begin
    Result := PAnsiChar(_Value);
  end;

  function TMGSVariable.IsConstant: boolean;
  begin
    Result := IsConstant;
  end;

  procedure TMGSVariable.SetValue(value: PAnsiChar);
  begin
    _SetValue(value);
  end;

  procedure TMGSVariable._SetValue(value: PAnsiChar);
  begin
    if _IsConstant then Exit;
    case _Type of
      mgsVT_Int:
          if IsIntVar(value) then
            _Value := value
          else
            _Value := defIntValue;
      mgsVT_Str:
          _Value := value;
      mgsVT_Real:
          if IsRealVar(value) then
            _Value := value
          else
            _Value := defRealValue;
      mgsVT_Boolean:
          if IsBoolVar(value) then
            _Value := value
          else
            _Value := defBooleanValue;
    end;
  end;

{ TMGSVariableScope }

function TMGSVariableScope.AddNewVariable(Name: PAnsiChar; VType: mgsVarType; Value: PAnsiChar; IsConstant: boolean; out Variable: IMGSVariable): cardinal;
  var _Variable: TMGSVariable;
      h: integer;
begin
  if GetVariable(Name, Variable) = MGS_E_OK then
  begin
    Result := MGS_E_NAME_EMPLOY;
    Exit;
  end;
  _Variable := TMGSVariable.Create(Name, VType, Value, IsConstant);
  if not _HandleStack.Pop(h) then
  begin
    _Variable._Handle := _VariableCount;
    inc(_VariableCount);
    SetLength(_VariableList, _VariableCount);
  end else
    _Variable._Handle := h;
  Variable := _Variable;
  _VariableList[_Variable._Handle] := Variable;
  Result := MGS_E_OK;
end;

constructor TMGSVariableScope.Create(Name: PAnsiChar);
begin
  inherited Create;
  _Name := Name;
  _HandleStack := TIntStack.Create;
end;

destructor TMGSVariableScope.Destroy;
  var i: cardinal;
begin
  if _VariableCount > 0 then
    for i := 0 to _VariableCount - 1 do
      _VariableList[i] := nil;
  _HandleStack.Free;
  inherited Destroy;
end;

function TMGSVariableScope.FreeVariable(Handle: cardinal): cardinal;
begin
  if Handle >= _VariableCount then
  begin
    Result := MGS_E_DATA_OUT;
    Exit;
  end;
  if _VariableList[Handle] = nil then
  begin
    Result := MGS_E_DATA_NULL;
    Exit;
  end;
  _VariableList[Handle] := nil;
  Result := MGS_E_OK;
end;

function TMGSVariableScope.GetName: PAnsiChar;
begin
  Result := PAnsiChar(_Name);
end;

function TMGSVariableScope.GetVariable(Handle: cardinal; out Variable: IMGSVariable): cardinal;
begin
  if Handle >= _VariableCount then
  begin
    Result := MGS_E_DATA_OUT;
    Exit;
  end;
  if _VariableList[Handle] = nil then
  begin
    Result := MGS_E_DATA_NULL;
    Exit;
  end;
  Variable := _VariableList[Handle];
  Result := MGS_E_OK;
end;

function TMGSVariableScope.GetVariable(Name: PAnsiChar; out Variable: IMGSVariable): cardinal;
  var i: integer;
begin
  for i := 0 to _VariableCount - 1 do
  begin
    if _VariableList[i] <> nil then
      if StrComp(_VariableList[i].GetName, Name) = 0 then
      begin
        Variable := _VariableList[i];
        Result := MGS_E_OK;
        Exit;
      end;
  end;
  Result := MGS_E_DATA_NOTFOUND;
end;

{ TMGSFunction }

  constructor TMGSFunction.Create(Name: PAnsiChar; FType: mgsFuncType; ParameterList: TMGSFuncParameterList; Count: integer; NativeFuncAddress: Pointer);
    var i: integer;
  begin
    inherited Create;
    _Name := GetMemory(StrLen(Name) + 1);
    StrCopy(_Name, Name);
    _Type := FType;
    if _Type = mgsFT_Return then
    begin
      SetLength(_ParameterList, count - 1);
      _ResultType := ParameterList[0]._VType;
      for i := 1 to count - 1 do
        _ParameterList[i - 1] := ParameterList[i];
    end else
      _ParameterList := ParameterList;
    _IsNative := true;
    _NativeFuncAddress := NativeFuncAddress;
    _VariableScope := TMGSVariableScope.Create('local');
  end;

  constructor TMGSFunction.Create (Name: PAnsiChar; FType: mgsFuncType; ParameterList: TMGSFuncParameterList; Count: integer; FunctionCode: PAnsiChar; ParentFunctionScope: IMGSFunctionScope);
    var i: integer;
  begin
    inherited Create;
    _Name := GetMemory(StrLen(Name) + 1);
    StrCopy(_Name, Name);
    _Type := FType;
    
    if _Type = mgsFT_Return then
    begin
      SetLength(_ParameterList, count - 1);
      _ResultType := ParameterList[0]._VType;
      for i := 1 to count - 1 do
        _ParameterList[i - 1] := ParameterList[i];
    end else
      _ParameterList := ParameterList;
      
    _IsNative := false;
    _VariableScope := TMGSVariableScope.Create('local');
    _FunctionCode := TMGSScript.Create(_Name, ParentFunctionScope, _VariableScope);
    _FunctionCode.Load(FunctionCode);
  end;

  destructor TMGSFunction.Destroy;
  begin
    _VariableScope := nil;
    inherited;
  end;

  function TMGSFunction.Execute(Values: TMGSFuncValueList; var returnValue: PAnsiChar): cardinal;
    var i, count: integer;
        iVar: IMGSVariable;
        //---
        proc: TMGSScriptProcedure;
        func: TMGSScriptFunction;
        //---
        ret: PAnsiChar;
  begin
    //    
    //      
    Result := MGS_E_OK;
    if not _IsNative then
    begin
      count := High(_ParameterList) + 1;
      for i := 0 to count - 1 do
      begin
        case _ParameterList[i]._PType of
          mgsPT_const:
            _VariableScope.AddNewVariable(_ParameterList[i]._Name, _ParameterList[i]._VType, MGSValueGetValue(@Values[i]) , false, iVar);
          mgsPT_ref:
            _VariableScope.AddNewVariable(_ParameterList[i]._Name, Values[i]._Ref.GetType, Values[i]._Ref.GetValue, false, iVar);
        end;
      end;
      if _Type = mgsFT_Return then
        _VariableScope.AddNewVariable('Result', _ResultType, nil, false, iVar);
    end;
    
    //
      // 
    if _IsNative then
    begin
      case _Type of
        mgsFT_Action:
          begin
            @proc := _NativeFuncAddress;
            proc(Values);
          end;
        mgsFT_Return:
          begin
            @func := _NativeFuncAddress;
            returnValue := func(Values);
          end;
      end;
    end else begin
      // 
      _FunctionCode.Execute;
      if (_Type = mgsFT_Return) then
      begin
        _VariableScope.GetVariable('Result', iVar);
        if (iVar <> nil) then
          returnValue := iVar.GetValue();  
      end;
    end;
  end;

  function TMGSFunction.GetHandle: cardinal;
  begin
    Result := _Handle;
  end;

  function TMGSFunction.GetName: PAnsiChar;
  begin
    Result := PAnsiChar(_Name);
  end;

  function TMGSFunction.GetReturnType: mgsVarType;
  begin
    Result := _ResultType;
  end;

  function TMGSFunction.GetType: mgsFuncType;
  begin
    Result := _Type;
  end;

  function TMGSFunction.VariableScope: IMGSVariableScope;
  begin
    Result := _VariableScope;
  end;

{ TMGSFunctionScope }

function TMGSFunctionScope.AddNewFunction(Name: PAnsiChar;
  FType: mgsFuncType; ParameterList: TMGSFuncParameterList; ParameterCount: integer;
  Address: Pointer; out Func: IMGSFunction): cardinal;
  //---
  var h: integer;
      _Func: TMGSFunction;
begin
  Result := MGS_E_OK;
  if GetFunction(Name, Func) = MGS_E_OK then
  begin
    Result := MGS_E_NAME_EMPLOY;
    Exit;
  end;

  _Func := TMGSFunction.Create(Name, FType, ParameterList, ParameterCount, Address);
  
  if not _HandleStack.Pop(h) then
  begin
    _Func._Handle := _FunctionCount;
    inc(_FunctionCount);
    SetLength(_FunctionList, _FunctionCount);
  end else
    _Func._Handle := h;
  _FunctionList[_Func._Handle] := _Func;
  Func := _Func;
end;

function TMGSFunctionScope.AddNewFunction(Name: PAnsiChar;
  FType: mgsFuncType; ParameterList: TMGSFuncParameterList; ParameterCount: integer;
  FunctionCode: PAnsiChar;
  ParentFunctionScope: IInterface; out Func: IMGSFunction): cardinal;
  //---
  var h: integer;
      _Func: TMGSFunction;
      _FuncScope: IMGSFunctionScope;
begin
  Result := MGS_E_OK;
  if (GetFunction(Name, Func) = MGS_E_OK) then
  begin
    Result := MGS_E_NAME_EMPLOY;
    Exit;
  end;
  
  ParentFunctionScope.QueryInterface(IMGSFunctionScope, _FuncScope);
  _Func := TMGSFunction.Create(Name, FType, ParameterList, ParameterCount, FunctionCode, _FuncScope);
  
  if not _HandleStack.Pop(h) then
  begin
    _Func._Handle := _FunctionCount;
    inc(_FunctionCount);
    SetLength(_FunctionList, _FunctionCount);
  end else
    _Func._Handle := h;
  _FunctionList[_Func._Handle] := _Func;
end;

constructor TMGSFunctionScope.Create(Name: PAnsiChar);
begin
  inherited Create;
  _Name := Name;
  _HandleStack := TIntStack.Create;
end;

destructor TMGSFunctionScope.Destroy;
  var i: integer;
begin
  for i := 0 to _FunctionCount - 1 do
    _FunctionList[i] := nil;
  _HandleStack.Free;
  inherited;
end;

function TMGSFunctionScope.GetFunction(Handle: cardinal; out Func: IMGSFunction): cardinal;
begin
  Result := MGS_E_OK;
  if Handle >= _FunctionCount then
  begin
    Result := MGS_E_DATA_OUT;
    Exit;
  end;
  if _FunctionList[Handle] = nil then
  begin
    Result := MGS_E_DATA_NULL;
    Exit;
  end;
  Func := _FunctionList[Handle];
end;

function TMGSFunctionScope.GetFunction(Name: PAnsiChar;
  out Func: IMGSFunction): cardinal;
  
  var i: integer;
      n: PAnsiChar;
begin
  Result := MGS_E_OK;
  for i := 0 to _FunctionCount - 1 do
    if _FunctionList[i] <> nil then
    begin
      if StrComp(_FunctionList[i].GetName, Name) = 0 then
      begin
        Func := _FunctionList[i];
        Exit;
      end;
    end;
  Result := MGS_E_DATA_NOTFOUND;
end;

function TMGSFunctionScope.GetName: PAnsiChar;
begin
  Result := _Name;
end;

{ TMGSScript }

constructor TMGSScript.Create(Name: PAnsiChar);
begin
  inherited Create();
  _VariableScope := TMGSVariableScope.Create('local');
  _FunctionScope := TMGSFunctionScope.Create('local');
  _CodeList := TStringList.Create;
  _Name := Name;
end;

constructor TMGSScript.Create(Name: PAnsiChar; FunctionScope: IMGSFunctionScope; VariableScope: IMGSVariableScope);
begin
  inherited Create();
  _Name := Name;
  _CodeList := TStringList.Create;
  _VariableScope := VariableScope;
  _FunctionScope := FunctionScope;
end;

destructor TMGSScript.Destroy;
begin
  _CodeList.Destroy;
  inherited;
end;


  procedure TMGSScript.SkipSpaceTokens(var pos: integer);
    var i: integer;
  begin
    for i := pos to _TokenCount - 1 do
      if not (_TokenList[i]._text[1] in [' ', #13, #10, #9]) then
      begin
        pos := i;
        Exit;
      end;
    pos := _TokenCount - 1;
  end;

  function TMGSScript.IsVariableDef(pos: integer): boolean;
  var token: TMGSToken;
begin
  Result := false;
  SkipSpaceTokens(pos);
  token := _TokenList[pos]; inc(pos);
  if not ((token._type = mgsTT_keyword) and (token._text = 'def')) then Exit;
  SkipSpaceTokens(pos);
  token := _TokenList[pos]; inc(pos);
  if (token._type = mgsTT_keyword) and (token._text = 'local') then
  begin
    SkipSpaceTokens(pos);
    token := _TokenList[pos]; inc(pos);
  end;
  if not ((token._type = mgsTT_delimeter) and (token._text = '[')) then Exit;
  SkipSpaceTokens(pos);
  token := _TokenList[pos]; inc(pos);
  if not ((token._type = mgsTT_keyword) and (
              (token._text = 'int') or
              (token._text = 'str') or
              (token._text = 'bool') or
              (token._text = 'real'))) then Exit;
  SkipSpaceTokens(pos);
  token := _TokenList[pos]; inc(pos);
  if not ((token._type = mgsTT_delimeter) and (token._text = ']')) then Exit;
  SkipSpaceTokens(pos);
  token := _TokenList[pos]; inc(pos);
  if token._type <> mgsTT_name then Exit;
  SkipSpaceTokens(pos);
  token := _TokenList[pos]; inc(pos);
  if (token._type = mgsTT_delimeter) and (token._text = ';') then
  begin
    Result := true;
    Exit;
  end;
  if (token._type = mgsTT_operator) and (token._text = '=') then
  begin
    Result := true;
    Exit;
  end;
end;

procedure TMGSScript.VariableDef(var pos: integer);
  var token: TMGSToken;
      varType: mgsVarType;
      varName: string;
      isLocal: boolean;
      eCode: cardinal;
      
      v: IMGSVariable;
begin
  isLocal := false;
  varName := '';

  SkipSpaceTokens(pos);
  token := _TokenList[pos]; inc(pos);
  if not ((token._type = mgsTT_keyword) and (token._text = 'def')) then Exit;
  SkipSpaceTokens(pos);
  token := _TokenList[pos]; inc(pos);
  if (token._type = mgsTT_keyword) and (token._text = 'local') then
  begin
    isLocal := true;
    SkipSpaceTokens(pos);
    token := _TokenList[pos]; inc(pos);
  end;
  
  if not ((token._type = mgsTT_delimeter) and (token._text = '[')) then Exit;
  
  SkipSpaceTokens(pos);
  token := _TokenList[pos]; inc(pos);
  if not ((token._type = mgsTT_keyword) and (
              (token._text = 'int') or
              (token._text = 'str') or
              (token._text = 'bool') or
              (token._text = 'real'))) then Exit;
              
  varType := GetVarType(token._text);
  SkipSpaceTokens(pos);
  
  token := _TokenList[pos]; inc(pos);
  if not ((token._type = mgsTT_delimeter) and (token._text = ']')) then Exit;
  
  SkipSpaceTokens(pos);
  token := _TokenList[pos]; inc(pos);
  if token._type <> mgsTT_name then Exit;
  
  varName := token._text;
  SkipSpaceTokens(pos);
  token := _TokenList[pos]; inc(pos);
  
  if (token._type = mgsTT_delimeter) and (token._text = ';') then
  begin
    if isLocal then
      eCode := _VariableScope.AddNewVariable(PAnsiChar(varName), varType, '', false, v)
    else
      eCode := GlobalVariableScope.AddNewVariable(PAnsiChar(varName), varType, '', false, v);
  end else if (token._type = mgsTT_operator) and (token._text = '=') then
  begin
    if isLocal then
      eCode := _VariableScope.AddNewVariable(PAnsiChar(varName), varType, GetValue(pos, false)._Value, false, v)
    else
      eCode := GlobalVariableScope.AddNewVariable(PAnsiChar(varName), varType, GetValue(pos, false)._Value, false, v);
  end;
  
  if eCode <> MGS_E_OK then
      MGSDebug('Can not create new variable with name "'+varName+'"' , eCode);
end;

function TMGSScript.IsVariableDelete(pos: integer): boolean;
  var token: TMGSToken;
begin
  Result := false;
  
  token := _TokenList[pos]; inc(pos);
  if (token._type <> mgsTT_keyword) or (token._text <> 'delete') then Exit;
  SkipSpaceTokens(pos);

  token := _TokenList[pos]; inc(pos);
  if (token._type = mgsTT_keyword) and (token._text = 'local') then
  begin
    SkipSpaceTokens(pos);
    token := _TokenList[pos]; inc(pos);
  end;

  if (token._type <> mgsTT_name) then Exit;
  
  Result := true;
end;

procedure TMGSScript.VariableDelete(var pos: integer);
  var token: TMGSToken;
      isLocal: boolean;
      outVar: IMGSVariable;
      eCode: cardinal;
begin
  token := _TokenList[pos]; inc(pos);
  isLocal := false;
  if (token._type <> mgsTT_keyword) or (token._text <> 'delete') then Exit;
  SkipSpaceTokens(pos);

  token := _TokenList[pos]; inc(pos);
  if (token._type = mgsTT_keyword) and (token._text = 'local') then
  begin
    isLocal := true;
    SkipSpaceTokens(pos);
    token := _TokenList[pos]; inc(pos);
  end;

  if (token._type <> mgsTT_name) then Exit;

  if (isLocal) then
  begin
    eCode := _VariableScope.GetVariable(PAnsiChar(token._text), outVar);
    if eCode <> MGS_E_OK then
    begin
      MGSDebug('Varialbe with name "'+token._text+'" not found in local scope', eCode);
      Exit;
    end;
    eCode := _VariableScope.FreeVariable(outVar.GetHandle());
  end else begin
    eCode := GlobalVariableScope.GetVariable(PAnsiChar(token._text), outVar);
    if eCode <> MGS_E_OK then
    begin
      MGSDebug('Varialbe with name "'+token._text+'" not found in global scope', eCode);
      Exit;
    end;
    eCode := GlobalVariableScope.FreeVariable(outVar.GetHandle());
  end;
  
  if eCode <> MGS_E_OK then
    MGSDebug('Can not delete variable with name "'+token._text+'"', eCode);
end;

function TMGSScript.IsElse(pos: integer): boolean;
  var token: TMGSToken;
begin
  token := _TokenList[pos]; inc(pos);
  Result := (token._type = mgsTT_keyword) and (token._text = 'else');    
end;

function TMGSScript.IsEndIF(pos: integer): boolean;
  var token: TMGSToken;
begin
  token := _TokenList[pos]; inc(pos);
  Result := (token._type = mgsTT_keyword) and (token._text = 'endif');
end;

function TMGSScript.IfAction(var pos: integer): boolean;
  var token: TMGSToken;
      i: integer;
      sCount: integer;
      
      val: TMGSValue;
begin
  Result := false;

  token := _TokenList[pos]; inc(pos);
  if (token._type <> mgsTT_keyword) or (token._text <> 'if') then Exit;
  SkipSpaceTokens(pos);
  token := _TokenList[pos]; inc(pos);
  if (token._type <> mgsTT_delimeter) or (token._text <> '(') then Exit;
  
  val := GetValue(pos, false);

  if val._Type <> mgsVT_Boolean then
  begin
    MGSDebug('Uncorrect IF statment result type', MGS_E_UNCORRECT_TYPE);
    Exit;
  end;

  Result := val._Value = 'True';

end;

function TMGSScript.IsIF(pos: integer): boolean;
  var token: TMGSToken;
      i: integer;
      sCount: integer;
begin
  Result := false;

  token := _TokenList[pos]; inc(pos);
  if (token._type <> mgsTT_keyword) or (token._text <> 'if') then Exit;
  SkipSpaceTokens(pos);
  token := _TokenList[pos]; inc(pos);
  if (token._type <> mgsTT_delimeter) or (token._text <> '(') then Exit;

  sCount := 1;
    
  for i := pos to _TokenCount - 1 do
    if token._type = mgsTT_delimeter then
    begin
      if token._text = '(' then inc(sCount)
      else if token._text = ')' then begin
        dec(sCount);
        if sCount < 0 then
          break;
      end;
    end;

  Result := true;
end;

  function TMGSScript.IsLoop(pos: integer): boolean;
    var token: TMGSToken;
  begin
    token := _TokenList[pos];
    Result := (token._type = mgsTT_keyword) and (token._text = 'loop');
  end;

  function TMGSScript.IsEndLoop(pos: integer): boolean;
    var token: TMGSToken;
  begin
    token := _TokenList[pos];
    Result := (token._type = mgsTT_keyword) and (token._text = 'endloop');
  end;

  function TMGSScript.IsExitWhen(pos: integer): boolean;
    var token: TMGSToken;
        i: integer;
        sCount: integer;
  begin
    Result := false;

    token := _TokenList[pos]; inc(pos);
    if (token._type <> mgsTT_keyword) or (token._text <> 'exitwhen') then Exit;
    SkipSpaceTokens(pos);
    token := _TokenList[pos]; inc(pos);
    if (token._type <> mgsTT_delimeter) or (token._text <> '(') then Exit;

    sCount := 1;
    for i := pos to _TokenCount - 1 do
      if token._type = mgsTT_delimeter then
      begin
        if token._text = '(' then inc(sCount)
        else if token._text = ')' then begin
          dec(sCount);
          if sCount < 0 then
            break;
        end;
      end;

    Result := true;
  end;

  function TMGSScript.LoopExitWhen(var pos: integer): boolean;
  var token: TMGSToken;
      i: integer;
      sCount: integer;
      val: TMGSValue;
  begin
    Result := false;

    token := _TokenList[pos]; inc(pos);
    if (token._type <> mgsTT_keyword) or (token._text <> 'exitwhen') then Exit;
    SkipSpaceTokens(pos);
    token := _TokenList[pos]; inc(pos);
    if (token._type <> mgsTT_delimeter) or (token._text <> '(') then Exit;

    val := GetValue(pos, false);
    
    if val._Type <> mgsVT_Boolean then
    begin
      MGSDebug('Uncorrect EXITWHEN statment result type', MGS_E_UNCORRECT_TYPE);
      Exit;
    end;
    Result := val._Value = 'True';
  end;

  function TMGSScript.IsAssignOperation(pos: integer): boolean;
    var token: TMGSToken;
  begin
    Result := false;
    
    token := _TokenList[pos]; inc(pos);
    if token._type <> mgsTT_name then Exit;

    SkipSpaceTokens(pos);

    token := _TokenList[pos]; inc(pos);
    if (token._type <> mgsTT_operator) or (token._text <> '=') then Exit;

    Result := true;    
  end;

  procedure TMGSScript.AssignOperation(var pos: integer);
    var token: TMGSToken;
        val: TMGSValue;
        variableName: string;
        variable: IMGSVariable;
  begin
    token := _TokenList[pos]; inc(pos);
    if token._type <> mgsTT_name then Exit;
    variableName := token._text;

    SkipSpaceTokens(pos);

    token := _TokenList[pos]; inc(pos);
    if (token._type <> mgsTT_operator) or (token._text <> '=') then Exit;

    if _VariableScope.GetVariable(PAnsiChar(variableName), variable) <> MGS_E_OK then
      if GlobalVariableScope.GetVariable(PAnsiChar(variableName), variable) <> MGS_E_OK then
      begin
        MGSDebug('Variable "'+variableName+'" not found');
        Exit;
      end;

    if variable = nil then
    begin
      MGSDebug('Variable data is NIL', MGS_E_DATA_NULL);
      Exit;
    end;
    
    val := GetValue(pos, false);
    variable.SetValue(val._Value);
  end;

  function TMGSScript.IsFunctionCall(pos: integer): boolean;
    var token: TMGSToken;
        i: integer;
        sCount: integer;
  begin
    Result := false;

    token := _TokenList[pos]; inc(pos);
    if token._type <> mgsTT_name then Exit;
    SkipSpaceTokens(pos);
    token := _TokenList[pos]; inc(pos);
    if token._text <> '(' then Exit;
    //Check parameters token
    sCount := 1;
    while (sCount > 0) and (pos < _TokenCount) do
    begin
      token := _TokenList[pos]; inc(pos);
      if (token._type = mgsTT_delimeter) then
      begin
        if (token._text = ')') then
          dec(sCount)
        else if (token._text = ')') then
          inc(sCount)
        else if (token._text = ';') then
          Exit;
      end;
    end;
    Result := true;
  end;

  function TMGSScript.GetValue(var pos: integer; subValue: boolean): TMGSValue;
    var i: integer;
        mathTree: PMathTree;
        child, tree: PMathTree;

        val: PMGSValue;
        hasLeft: boolean;
        left: integer;
        token: TMGSToken;

        sCount: integer;
        opMass: integer;

        function GetOperatorMass(str: string): integer;
        begin
          Result := 0;
          if (str = '+') or (str = '-') then
            Result := 0
          else if (str = '*') or (str = '/') then
            Result := 1
          else if (str = '<') or (str = '<=') or (str = '>') or (str = '>=') or (str = '==') or (str = '!=') or (str = 'not') then
            Result := 2;
        end;

        //function for calculate math tree
        function CalculateTree(base: PMathTree): PMGSValue;
          var operator: mgsMathOperator;
              lValue, rValue: PMGSValue;

            function GetOpType(text: string): mgsMathOperator;
            begin
            //Std math
              if text = '+' then
                Result := mgsMT_Plus
              else if text = '-' then
                Result := mgsMT_Minus
              else if text = '*' then
                Result := mgsMT_Mul
              else if text = '/' then
                Result := mgsMT_Div
            //Bool math
              else if text = 'or' then
                Result := mgsMT_Or
              else if text = 'and' then
                Result := MgsMT_And
              else if text = 'not' then
                Result := MgsMT_Not
            //Compress math
              else if text = '>' then
                Result := mgsMT_Great
              else if text = '>=' then
                Result := mgsMT_GreatEqual
              else if text = '<' then
                Result := mgsMT_Less
              else if text = '<=' then
                Result := mgsMT_LessEqual
              else if text = '==' then
                Result := mgsMT_Equal
              else if text = '!=' then
                Result := mgsMT_NonEqual;
            end;
            
        begin
          case base._Type of
            mathTT_value:
              begin
                New(Result);
                MGSValueMove(Result, @base._Value);
              end;
            mathTT_result:
              if base._Left <> nil then
                Result := CalculateTree(base._Left)
              else if base._Right <> nil then
                Result := CalculateTree(base._Right);
            mathTT_operator:
              begin
                operator := GetOpType(_TokenList[base._TokenPos]._text);
                New(lValue);
                New(rValue);
                if base._Left <> nil then
                  MGSValueMove( lValue, CalculateTree(base._Left) );
                if base._Right <> nil then
                  MGSValueMove( rValue, CalculateTree(base._Right) );
                New(Result);
                MathValues(lValue, rValue, operator, Result); 
              end;  
          end;
        end;

        function GetVariableValue(str: string): TMGSValue;
          var v: IMGSVariable;
              len: integer;
        begin
          str := Copy(str, 2, Length(str)-1);
          if _VariableScope.GetVariable(PAnsiChar(str), v) <> MGS_E_OK then
            if GlobalVariableScope.GetVariable(PAnsiChar(str), v) <> MGS_E_OK then
              Exit;
          Result.isConst := true;
          Result._Type := v.GetType;
          len := StrLen(v.GetValue);
          Result._Value := GetMemory(len+1);
          StrMove(Result._Value, v.GetValue, len+1);          
        end;

  begin
    Result.isConst := true;
    i := pos;
    sCount := 1;
    mathTree := TMathTree_Create();
    mathTree._Type := mathTT_result;
    mathTree._Text := 'Result';
    left := pos;
    hasLeft := false;

    child := mathTree;
    while ( i < _TokenCount ) do
    begin
      token := _TokenList[i];

      if (token._type = mgsTT_delimeter) then
      begin
        if token._text = '{' then
        begin
          inc(sCount);
          tree := TMathTree_Create();
          tree._Type := mathTT_value;
          opMass := i+1;
          tree._Value := Self.GetValue(opMass, true);
          tree._Text := 'sub';
          i := opMass - 1;
          TMathTree_AddChild(child, tree, not hasLeft);
          if hasLeft then
            hasLeft := false;
          child := tree;
        end else 
        if token._text = '}' then
        begin
          dec(sCount);
          if (sCount <= 0) then
            break;
        end else
        if token._text[1] in [#13, #10, ')', ';',','] then
        begin
          break;
        end;
      end else
      if (token._type = mgsTT_intconst) or
         (token._type = mgsTT_strconst) or
         (token._type = mgsTT_boolconst) or
         (token._type = mgsTT_name) then
      begin
        tree := TMathTree_Create();
        tree._Type := mathTT_value;
        tree._Value.isConst := true;
        tree._Value._Value := PAnsiChar(token._text);
        if (token._type = mgsTT_intConst) then
        begin
          if IsIntVar(token._text) then
            tree._Value._Type := mgsVT_Int
          else
            tree._Value._Type := mgsVT_Real;
        end
        else if (token._type = mgsTT_strConst) then
        begin
          tree._Value._Type := mgsVT_Str;
        end
        else if (token._type = mgsTT_boolConst) then
        begin
          tree._Value._Type := mgsVT_Boolean;
        end
        else if (token._type = mgsTT_name) then
        begin
          if token._text[1] = '@' then
            tree._Value := GetVariableValue(token._text)
          else if IsFunctionCall(i) then
            tree._Value := FunctionCall(i);
        end;

        tree._Text := token._text;
        if (child._Parent <> nil) and (child._Parent._Type = mathTT_operator) and (child._Type = mathTT_operator) and (child._Text = '-') then
        begin
          TMathTree_AddChild(child, tree, false);
          child := tree;
          hasLeft := false;
        end else
        if (child._Type = mathTT_operator) and (child._Text = 'not') then
        begin
          TMathTree_AddChild(child, tree, false);
          child := tree;
          hasLeft := false;
        end else begin
          TMathTree_AddChild(child, tree, not hasLeft);
          if (hasLeft) then
          begin
            hasLeft := false;
            child := tree;
          end;
        end;
      end else
      if (token._type = mgsTT_operator) then
      begin
        opMass := GetOperatorMass(token._text);
        if (not ((child._Type = mathTT_operator) and (token._text = '-'))) and (token._text <> 'not') then
          tree := child._Parent
        else
          tree := child;

        if (tree <> nil) then
        begin

          while (tree <> nil) and (tree._Type <> mathTT_result) and ((tree._Type = mathTT_operator) and (tree._Text = '-') and (tree._Left = nil)) do
            tree := tree._Parent;

          if (token._Text <> 'not') then          
            while (tree <> nil) and (tree._Type <> mathTT_result) and (tree._Type = mathTT_operator) and (tree._Text = 'not') do
              tree := tree._Parent; 

          if (not (child._Type = mathTT_operator) and (token._text = '-')) and (opMass = 0) then
          begin
            while (tree <> nil) and ((tree._Type <> mathTT_result) or (tree._Type <> mathTT_operator)) and (opMass <> GetOperatorMass(_TokenList[tree._TokenPos]._text)) do
              tree := tree._Parent;
          end;

          if (tree._Type = mathTT_operator) or (tree._Type = mathTT_result) then
            child := tree;
          tree := TMathTree_Create();
          tree._Type := mathTT_operator;
          tree._TokenPos := i;
          tree._Text := token._text;
          if (child._Type = mathTT_operator) or (child._Type = mathTT_result) then
            if ((child._Type = mathTT_operator) and (token._text = '-')) or ((child._Type = mathTT_operator) and (child._Text = 'not')) or (child._TokenPos >= 0) and (opMass > GetOperatorMass(_TokenList[child._TokenPos]._text)) then
              TMathTree_InsertRight(child, tree)
            else
              TMathTree_InsertLeft(child, tree)
          else
            TMathTree_InsertRight(child._Parent, tree);
        end else begin
          tree := TMathTree_Create();
          tree._Type := mathTT_operator;
          tree._TokenPos := i;
          tree._Text := token._text;
          TMathTree_InsertLeft(child, tree);
        end;
        child := tree;
        hasLeft := true;
      end;

      inc(i);
    end;
    pos := i;

    //ShowMessage(TMathTree_Draw(mathTree));
    //calc tree
    val := CalculateTree(mathTree);
    MGSValueMove(@Result, val);
  end;

  function TMGSScript.IsFunctionDef(pos: integer): boolean;
    var token: TMGSToken;
  begin
    Result := false;
    
    token := _TokenList[pos]; inc(pos);
    if (token._type <> mgsTT_keyword) or (token._text <> 'function') then Exit;
    SkipSpaceTokens(pos);

    token := _TokenList[pos]; inc(pos);
    SkipSpaceTokens(pos);

    //  global
    if (token._type <> mgsTT_delimeter) and (token._type <> mgsTT_name) then
    begin
      if (token._type <> mgsTT_keyword) or (token._text <> 'global') then
        Exit;

      token := _TokenList[pos]; inc(pos);
      SkipSpaceTokens(pos);
    end;


    if (token._type = mgsTT_delimeter) and (token._text = '[') then
    begin
      token := _TokenList[pos]; inc(pos);
      if (token._type <> mgsTT_keyword) or
            ((token._text <> 'int') and
            (token._text <> 'real') and
            (token._text <> 'bool') and
            (token._text <> 'str')) then
        Exit;
      SkipSpaceTokens(pos);
      
      token := _TokenList[pos]; inc(pos);
      if (token._type <> mgsTT_delimeter) or (token._text <> ']') then Exit;
      SkipSpaceTokens(pos);

      token := _TokenList[pos]; inc(pos);
      SkipSpaceTokens(pos);
    end;

    if (token._type <> mgsTT_name) then Exit;

    token := _TokenList[pos]; inc(pos);
    if (token._type <> mgsTT_delimeter) or (token._text <> '(') then Exit;
    SkipSpaceTokens(pos);

    while (true) do
    begin
      token := _TokenList[pos]; inc(pos);
      if (token._type <> mgsTT_delimeter) or (token._text <> '[') then
      begin
        if (token._type = mgsTT_delimeter) and (token._text = ')') then
          Result := true;
        Exit;
      end;
      SkipSpaceTokens(pos);

      token := _TokenList[pos]; inc(pos);
      if (token._type <> mgsTT_keyword) or (
                (token._text <> 'int') and
                (token._text <> 'real') and
                (token._text <> 'bool') and
                (token._text <> 'str')) then
          Exit;
      SkipSpaceTokens(pos);

      token := _TokenList[pos]; inc(pos);
      if (token._type <> mgsTT_delimeter) or (token._text <> ']') then Exit;
      SkipSpaceTokens(pos);

      token := _TokenList[pos]; inc(pos);
      if (token._type <> mgsTT_name) then Exit;
      SkipSpaceTokens(pos);

      token := _TokenList[pos]; inc(pos);
      if (token._type <> mgsTT_delimeter) or (token._text <> ',') then
      begin
        if (token._type = mgsTT_delimeter) and (token._text = ')') then
          Result := true;
        Exit;
      end;
      SkipSpaceTokens(pos);
    end;
    
    Result := true;
  end;

  procedure TMGSScript.FunctionDef(var pos: integer);
    var token: TMGSToken;
        funcType: mgsFuncType;
        funcName: string;
        funcCode: string;
        parameterList: TMGSFuncParameterList;
        parameterCount: integer;

        func: IMGSFunction;

        pType: mgsVarType;
        idx: integer;
        isGlobal: boolean;
  begin
    funcType := mgsFT_Action;
    isGlobal := false;
    parameterCount := 0;
    funcName := '';
    funcCode := '';

    token := _TokenList[pos]; inc(pos);
    if (token._type <> mgsTT_keyword) or (token._text <> 'function') then Exit;
    SkipSpaceTokens(pos);

    token := _TokenList[pos]; inc(pos);
    SkipSpaceTokens(pos);

    //  global
    if (token._type <> mgsTT_delimeter) and (token._type <> mgsTT_name) then
    begin
      if (token._type <> mgsTT_keyword) or (token._text <> 'global') then
        Exit;

      isGlobal := true;
      token := _TokenList[pos]; inc(pos);
      SkipSpaceTokens(pos);
    end;

    if (token._type = mgsTT_delimeter) and (token._text = '[') then
    begin
      token := _TokenList[pos]; inc(pos);
      if (token._type <> mgsTT_keyword) or
            ((token._text <> 'int') and
            (token._text <> 'real') and
            (token._text <> 'bool') and
            (token._text <> 'str')) then
        Exit;
      SkipSpaceTokens(pos);

      funcType := mgsFT_Return;
      inc(parameterCount);
      SetLength(parameterList, parameterCount);
      parameterList[0]._PType := mgsPT_const;
      parameterList[0]._VType := GetVarType(token._text);
      parameterList[0]._Name  := 'Result';

      token := _TokenList[pos]; inc(pos);
      if (token._type <> mgsTT_delimeter) or (token._text <> ']') then Exit;
      SkipSpaceTokens(pos);

      token := _TokenList[pos]; inc(pos);
      SkipSpaceTokens(pos);
    end;

    if (token._type <> mgsTT_name) then Exit;
    funcName := token._text;

    token := _TokenList[pos]; inc(pos);
    if (token._type <> mgsTT_delimeter) or (token._text <> '(') then Exit;
    SkipSpaceTokens(pos);

    while (true) do
    begin
      token := _TokenList[pos]; inc(pos);
      if (token._type <> mgsTT_delimeter) or (token._text <> '[') then
      begin
        if (token._type = mgsTT_delimeter) and (token._text = ')') then
          break;
        Exit;
      end;      
      SkipSpaceTokens(pos);

      token := _TokenList[pos]; inc(pos);
      if (token._type <> mgsTT_keyword) or (
                (token._text <> 'int') and
                (token._text <> 'real') and
                (token._text <> 'bool') and
                (token._text <> 'str')) then
          Exit;

      SkipSpaceTokens(pos);
      pType := GetVarType(token._text);

      token := _TokenList[pos]; inc(pos);
      if (token._type <> mgsTT_delimeter) or (token._text <> ']') then Exit;
      SkipSpaceTokens(pos);

      token := _TokenList[pos]; inc(pos);
      if (token._type <> mgsTT_name) then Exit;
      SkipSpaceTokens(pos);

      idx := parameterCount;
      inc(parameterCount);
      SetLength(parameterList, parameterCount);
      parameterList[idx]._PType := mgsPT_const;
      parameterList[idx]._VType := pType;
      parameterList[idx]._Name  := PAnsiChar(token._text);

      token := _TokenList[pos]; inc(pos);
      if (token._type <> mgsTT_delimeter) or (token._text <> ',') then
      begin
        if (token._type = mgsTT_delimeter) and (token._text = ')') then
          break;
        Exit;
      end;
      SkipSpaceTokens(pos);
    end;
    
    //Read core
    while (true) do
    begin
      token := _TokenList[pos]; inc(pos);
      if (token._type = mgsTT_keyword) and (token._text = 'endfunction') then break;
      if token._type = mgsTT_strconst then
        funcCode := funcCode + '"' + token._text + '"'
      else
        funcCode := funcCode + token._text;
    end;

    funcCode := 'InitEnd'+#13#10 + funcCode + #13#10 + 'Stop;';

    if isGlobal then
      idx := GlobalFunctionScope.AddNewFunction(PAnsiChar(funcName), funcType, parameterList, parameterCount, PAnsiChar(funcCode), _FunctionScope, func)
    else
      idx := _FunctionScope.AddNewFunction(PAnsiChar(funcName), funcType, parameterList, parameterCount, PAnsiChar(funcCode), _FunctionScope, func);
      
    if idx <> MGS_E_OK then
    begin
      MGSDebug('Can not create new function with name "'+funcName+'"', idx);
      Exit;
    end;
  end;

  function TMGSScript.FunctionCall(var pos: integer): TMGSValue;
    var token: TMGSToken;
        funcName: string;

        ret: PAnsiChar;
        func: IMGSFunction;
        params: TMGSFuncValueList;
        pcount: integer;
        scount: integer;
        oldpos: integer;

        eCode: cardinal;
  begin
    token := _TokenList[pos]; inc(pos);
    funcName := token._text;
    if _FunctionScope.GetFunction(PAnsiChar(funcName), func) <> MGS_E_OK then
      if GlobalFunctionScope.GetFunction(PAnsiChar(funcName), func) <> MGS_E_OK then
      begin
        MGSDebug('Function "'+funcName+'" is not found', MGS_E_DATA_NULL);
        Exit;
      end;
      
    pcount := 0;
    scount := 0;
    while (pos < _TokenCount) do
    begin
      token := _TokenList[pos];
      if (token._type = mgsTT_delimeter) then
      begin
        if (token._text = ')') then
        begin
          dec(scount); inc(pos);
          if (scount <= 0) then break;
          Continue;
        end else if (token._text = '(') then
        begin
          inc(scount); inc(pos);
          Continue;
        end;
      end;
      if (scount = 1) then
      begin
        inc(pcount);
        SetLength(params, pcount);
        params[pcount-1] := GetValue(pos, false);

        token := _TokenList[pos]; inc(pos);
        if (token._type <> mgsTT_delimeter) or (token._text <> ',') then
        begin
          if (token._type = mgsTT_delimeter) and (token._text = ')') then
            break;
        end;
        SkipSpaceTokens(pos);
      end;
    end;
    dec(pos);

    eCode := func.Execute(params, ret);
    if eCode <> MGS_E_OK then
    begin
      MGSDebug('Function execute error', eCode);
      Exit;
    end;

    if func.GetType = mgsFT_Return then
    begin
      Result.isConst := true;
      Result._Type := func.GetReturnType;
      Result._Value := ret;
    end;

  end;

  function TMGSScript.IsScriptStart(pos: integer): boolean;
    var token: TMGSToken;
  begin
    token := _TokenList[pos];
    Result := (token._type = mgsTT_keyword) and (token._text = 'InitEnd');
  end;

  function TMGSScript.IsScriptStop(pos: integer): boolean;
    var token: TMGSToken;
  begin
    token := _TokenList[pos];
    Result := (token._type = mgsTT_keyword) and (token._text = 'Stop');
  end;


procedure TMGSScript.Execute;
  var tokenPos, j: integer;
      localIntVal: array [0..9] of integer;
      
      _ifValueList: array of boolean;
      _ifPosList: array of boolean;   //true - in then, false - in else
      _ifSkip: boolean;
      _ifCurrentPosition: integer;
      _ifValueCount: integer;

      _loopValueList: array of boolean;
      _loopStartList: array of integer;
      _loopValueCount: integer;
      _loopCurrentPosition: integer;

      scriptStart: boolean;
begin
  tokenPos := 0;
  _ifValueCount := 0;
  _ifCurrentPosition := -1;
  _ifSkip := false;

  _loopValueCount := 0;
  _loopCurrentPosition := -1;

  scriptStart := false;

  while tokenPos < _TokenCount do
  begin
    SkipSpaceTokens(tokenPos);
    j := tokenPos;
    //*****************
    //  InitEnd
    if (not scriptStart) then
    begin
      if IsScriptStart(tokenPos) then
        scriptStart := true
      //************************
      //  function [int] name ([str] param1, ...)
      //    ...
      //  endfunction
      else if IsFunctionDef(tokenPos) then
        FunctionDef(tokenPos);

      inc(tokenPos);
      continue;
    end;      
    //*****************
    //  if (condition)
    if IsIF(tokenPos) then
    begin
      localIntVal[0]:= _ifValueCount;
      inc(_ifValueCount);
      SetLength(_ifValueList, _ifValueCount);
      SetLength(_ifPosList, _ifValueCount);
      _ifPosList[localIntVal[0]] := true;
      if (_ifCurrentPosition < 0) or (_ifPosList[_ifCurrentPosition] = _ifValueList[_ifCurrentPosition]) then
      begin
        _ifCurrentPosition := localIntVal[0];
        _ifValueList[localIntVal[0]] := IfAction(tokenPos);
      end;
    end
    //******************
    //  else
    else if IsElse(tokenPos) and (_ifValueCount > 0) then
        _ifPosList[_ifValueCount - 1] := not _ifPosList[_ifValueCount - 1]
    //******************
    //  endif
    else if IsEndIF(tokenPos) then
    begin
      dec(_ifValueCount);
      if _ifCurrentPosition >= _ifValueCount then
        _ifCurrentPosition := _ifValueCount - 1;
    end
      
    else if (_ifCurrentPosition < 0) or (_ifPosList[_ifCurrentPosition] = _ifValueList[_ifCurrentPosition]) then
    begin
      //***************
      //  loop
      if IsLoop(tokenPos) then
      begin
        localIntVal[0] := _loopValueCount;
        inc(_loopValueCount);
        SetLength(_loopValueList, _loopValueCount);
        SetLength(_loopStartList, _loopValueCount);
        _loopValueList[localIntVal[0]] := true;
        _loopStartList[localIntVal[0]] := tokenPos;
        if (_loopCurrentPosition < 0) or (_loopValueList[_loopCurrentPosition]) then
          _loopCurrentPosition := localIntVal[0];
      end
      else if IsEndLoop(tokenPos) then
      begin
        if (_loopCurrentPosition >= 0) and (_loopValueList[_loopCurrentPosition]) then
          tokenPos := _loopStartList[_loopCurrentPosition]
        else begin
          dec(_loopValueCount);
          if (_loopCurrentPosition >= _loopValueCount) then
            _loopCurrentPosition := _loopValueCount - 1;
        end;
      end
      else if (_loopCurrentPosition < 0) or (_loopValueList[_loopCurrentPosition]) then
      begin
        //***************
        //  exitwhen (condition)
        if IsExitWhen(tokenPos) and (_loopCurrentPosition >= 0) then
          _loopValueList[_loopCurrentPosition] := not LoopExitWhen(tokenPos)
        //***************
        //  name = value;
        else if IsAssignOperation(tokenPos) then
          AssignOperation(tokenPos)
        //***************
        //  def local [type] name = value;
        //  def [type] name = value;
        else if IsVariableDef(tokenPos) then
          VariableDef(tokenPos)
        //***************
        //  delete local name;
        //  delete name;
        else if IsVariableDelete(tokenPos) then
          VariableDelete(tokenPos)
        //***************
        //  Stop
        else if IsScriptStop(tokenPos) then
          break
        //***************
        //  function(parameters)
        else if IsFunctionCall(tokenPos) then
          FunctionCall(tokenPos);
      end;//end in "LOOP"
    end;//end in "IF"
      
    inc(tokenPos);
  end;
end;

function TMGSScript.GetName: PAnsiChar;
begin
  Result := _Name;
end;

procedure TMGSScript.Load(text: PAnsiChar);
begin
  _CodeList.Text := text;
  ParseTokens;
end;

procedure TMGSScript.ParseTokens();
  var i,j: integer;
      c: char;
      code, line, word: string;

  procedure SkipSpace(text: string; var spos: integer);
    var i : integer;
  begin
    for i := spos to Length(text) do
      if not (text[i] in [' ', #13, #10]) then
      begin
        spos := i - 1;
        Exit;
      end;
    spos := Length(text);
  end;

  procedure SkipComment(text: string; var spos: integer);
    var i : integer;
  begin
    for i := spos to Length(text) do
      if text[i] in [#13] then
      begin
        spos := i;
        Exit;
      end;
    spos := Length(text);
  end;

  function ReadString(text: string; var spos: integer): string;
    var i: integer;
  begin
    Result := '';
    for i := spos to Length(text) do
      if text[i] = '"' then
      begin
        spos := i;
        Exit;
      end else
        Result := Result + text[i];
    spos := Length(text);
  end;

  procedure AddToken(text: string; t: TMGSTokenType);
    var id: integer;
  begin
    id := _TokenCount;
    inc(_TokenCount);
    SetLength(_TokenList, _TokenCount);
    _TokenList[id]._text := text;
    _TokenList[id]._type := t;
  end;

  function IsOperator (word: string): boolean;
    var i: integer;
  begin
    Result := true;
    for i := 0 to operator_count-1 do
      if (word = tokentype_operator[i]) then
        Exit;
    Result := false;
  end;

  function IsKeyword (word: string): boolean;
    var i: integer;
  begin
    Result := true;
    for i := 0 to keyword_count-1 do
      if (word = tokentype_keyword[i]) then
        Exit;
    Result := false;
  end;

  procedure TokenToMsg(token: TMGSToken);
    var o: string;
  begin
    o := 'Text: "'+token._text+'"';
    o := o + #13#10 + 'Type: ';
    case token._type of
      mgsTT_keyword: o := o + 'keyword';
      mgsTT_operator: o := o + 'operator';
      mgsTT_name: o := o + 'name';
      mgsTT_delimeter: o := o + 'delimeter';
      mgsTT_intconst: o := o + 'int const';
      mgsTT_strconst: o := o + 'str const';
      mgsTT_boolconst: o := o + 'bool const';
    else
      o := o + 'unknown';
    end;
    ShowMessage(o);
  end;

begin
  code := _CodeList.Text;
  word := '';
  i := 1;
  while i <= Length(code) do
  begin
    c := code[i];
    //Read string
    if (c = '"') then
    begin
      inc(i);
      line := ReadString(code, i);
      AddToken(line, mgsTT_strconst);
      inc(i);
      Continue;  
    end;
    //Skip Comment
    if (c = '/') and (code[i+1] = '/') then
    begin
      SkipComment(code, i);
      inc(i);
      continue;
    end;
    if (not IsRealVar(word+c)) and (Pos(c, tokentype_delimeter) > 0) then
    begin
      //Add token
      if word <> '' then
      begin
        if IsOperator(word) then
          AddToken(word, mgsTT_operator)
        else if IsKeyword(word) then
          AddToken(word, mgsTT_keyword)
        else
          if (IsRealVar(word)) then
            AddToken(word, mgsTT_intconst)
          else if (IsBoolVar(word)) then
            AddToken(word, mgsTT_boolConst)
          else
            AddToken(word, mgsTT_name);
      end;
      word := '';

      //Skip space tokens
      if c in [' ', #13, #10] then
      begin
        SkipSpace(code, i);
        AddToken(' ', mgsTT_delimeter);
      end else
        AddToken(c, mgsTT_delimeter);
    end else
    begin
      if word = '' then
      begin
          if IsOperator(c + code[i+1]) then
          begin
            AddToken(c + code[i+1], mgsTT_operator);
            word := ''; i := i + 2;
            Continue;
          end else if IsOperator(c) then
          begin
            AddToken(c, mgsTT_operator);
            word := ''; inc(i);
            Continue;
          end;
          word := word + c;
      end else begin
        if IsRealVar(word) then
        begin
          if c in ['0'..'9', '.', '@'] then
            word := word + c
          else begin
            AddToken(word, mgsTT_intconst);
            word := '';
            if IsOperator(c+code[i+1]) then
            begin
              AddToken(c + code[i+1], mgsTT_operator);
              inc(i);
            end
            else if IsOperator(c) then
              AddToken(c, mgsTT_operator)
            else
              word := c;
          end;
        end else
          if c in ['a'..'z', 'A'..'Z', '_'] then
            word := word + c
          else begin
            if IsOperator(word) then
              AddToken(word, mgsTT_operator)
            else if IsKeyword(word) then
              AddToken(word, mgsTT_keyword)
            else if IsBoolVar(word) then
              AddToken(word, mgsTT_boolconst)
            else
              AddToken(word, mgsTT_name);
            word := c;
          end;
      end;
    end;
      
    inc(i);
  end;

  //for i := 0 to _TokenCount - 1 do TokenToMsg(_TokenList[i]);

end;









procedure MGSFunc_Debug(Values: TMGSFuncValueList);
begin
  ShowMessage(MGSValueGetValue(@Values[0]));
end;

function MGSFunc_GetRndValue(Values: TMGSFuncValueList): PAnsiChar;
  var s: string;
begin
  s := RealToStr(Random);
  Result := GetMemory(Length(s)+1);
  StrMove(Result, PAnsiChar(s), Length(s));
end;



procedure Init;
  var parList: TMGSFuncParameterList;
      func: IMGSFunction;
begin
  GlobalFunctionScope := TMGSFunctionScope.Create('Global');
  GlobalVariableScope := TMGSVariableScope.Create('Global');

  //Debug
  SetLength(parList, 1);
  parList[0]._PType := mgsPT_const;
  parList[0]._VType := mgsVT_Str;
  parList[0]._Name  := 'message';
  GlobalFunctionScope.AddNewFunction('Debug', mgsFT_Action, parList, 1, @MGSFunc_Debug, func);
  //Random
  SetLength(parList, 1);
  parList[0]._PType := mgsPT_const;
  parList[0]._VType := mgsVT_Real;
  parList[0]._Name := 'Result';
  GlobalFunctionScope.AddNewFunction('GetRndValue', mgsFT_Return, parList, 1, @MGSFunc_GetRndValue, func);
end;


end.
